@@ -45,7 +45,20 @@ struct gatt_db {
 	struct queue *services;
 };
 
+struct pending_read {
+	unsigned int id;
+	gatt_db_attribute_read_t func;
+	void *user_data;
+};
+
+struct pending_write {
+	unsigned int id;
+	gatt_db_attribute_write_t func;
+	void *user_data;
+};
+
 struct gatt_db_attribute {
+	struct gatt_db_service *service;
 	uint16_t handle;
 	bt_uuid_t uuid;
 	uint32_t permissions;
@@ -55,6 +68,12 @@ struct gatt_db_attribute {
 	gatt_db_read_t read_func;
 	gatt_db_write_t write_func;
 	void *user_data;
+
+	unsigned int read_id;
+	struct queue *pending_reads;
+
+	unsigned int write_id;
+	struct queue *pending_writes;
 };
 
 struct gatt_db_service {
@@ -63,14 +82,21 @@ struct gatt_db_service {
 	struct gatt_db_attribute **attributes;
 };
 
-static bool match_service_by_handle(const void *data, const void *user_data)
+static void attribute_destroy(struct gatt_db_attribute *attribute)
 {
-	const struct gatt_db_service *service = data;
+	/* Attribute was not initialized by user */
+	if (!attribute)
+		return;
 
-	return service->attributes[0]->handle == PTR_TO_INT(user_data);
+	queue_destroy(attribute->pending_reads, free);
+	queue_destroy(attribute->pending_writes, free);
+
+	free(attribute->value);
+	free(attribute);
 }
 
-static struct gatt_db_attribute *new_attribute(const bt_uuid_t *type,
+static struct gatt_db_attribute *new_attribute(struct gatt_db_service *service,
+							const bt_uuid_t *type,
 							const uint8_t *val,
 							uint16_t len)
 {
@@ -80,29 +106,30 @@ static struct gatt_db_attribute *new_attribute(const bt_uuid_t *type,
 	if (!attribute)
 		return NULL;
 
+	attribute->service = service;
 	attribute->uuid = *type;
 	attribute->value_len = len;
 	if (len) {
 		attribute->value = malloc0(len);
-		if (!attribute->value) {
-			free(attribute);
-			return NULL;
-		}
+		if (!attribute->value)
+			goto failed;
 
 		memcpy(attribute->value, val, len);
 	}
 
+	attribute->pending_reads = queue_new();
+	if (!attribute->pending_reads)
+		goto failed;
+
+	attribute->pending_writes = queue_new();
+	if (!attribute->pending_reads)
+		goto failed;
+
 	return attribute;
-}
 
-static void attribute_destroy(struct gatt_db_attribute *attribute)
-{
-	/* Attribute was not initialized by user */
-	if (!attribute)
-		return;
-
-	free(attribute->value);
-	free(attribute);
+failed:
+	attribute_destroy(attribute);
+	return NULL;
 }
 
 struct gatt_db *gatt_db_new(void)
@@ -138,6 +165,9 @@ static void gatt_db_service_destroy(void *data)
 
 void gatt_db_destroy(struct gatt_db *db)
 {
+	if (!db)
+		return;
+
 	queue_destroy(db->services, gatt_db_service_destroy);
 	free(db);
 }
@@ -156,8 +186,10 @@ static int uuid_to_le(const bt_uuid_t *uuid, uint8_t *dst)
 	return bt_uuid_len(&uuid128);
 }
 
-uint16_t gatt_db_add_service(struct gatt_db *db, const bt_uuid_t *uuid,
-					bool primary, uint16_t num_handles)
+struct gatt_db_attribute *gatt_db_add_service(struct gatt_db *db,
+						const bt_uuid_t *uuid,
+						bool primary,
+						uint16_t num_handles)
 {
 	struct gatt_db_service *service;
 	const bt_uuid_t *type;
@@ -184,7 +216,7 @@ uint16_t gatt_db_add_service(struct gatt_db *db, const bt_uuid_t *uuid,
 
 	len = uuid_to_le(uuid, value);
 
-	service->attributes[0] = new_attribute(type, value, len);
+	service->attributes[0] = new_attribute(service, type, value, len);
 	if (!service->attributes[0]) {
 		gatt_db_service_destroy(service);
 		return 0;
@@ -203,18 +235,21 @@ uint16_t gatt_db_add_service(struct gatt_db *db, const bt_uuid_t *uuid,
 	db->next_handle += num_handles;
 	service->num_handles = num_handles;
 
-	return service->attributes[0]->handle;
+	return service->attributes[0];
 }
 
-bool gatt_db_remove_service(struct gatt_db *db, uint16_t handle)
+bool gatt_db_remove_service(struct gatt_db *db,
+					struct gatt_db_attribute *attrib)
 {
 	struct gatt_db_service *service;
 
-	service = queue_remove_if(db->services, match_service_by_handle,
-							INT_TO_PTR(handle));
-	if (!service)
+	if (!db || !attrib)
 		return false;
 
+	service = attrib->service;
+
+	queue_remove(db->services, service);
+
 	gatt_db_service_destroy(service);
 
 	return true;
@@ -239,8 +274,8 @@ static uint16_t get_handle_at_index(struct gatt_db_service *service,
 	return service->attributes[index]->handle;
 }
 
-static uint16_t update_attribute_handle(struct gatt_db_service *service,
-								int index)
+static struct gatt_db_attribute *
+attribute_update(struct gatt_db_service *service, int index)
 {
 	uint16_t previous_handle;
 
@@ -250,7 +285,7 @@ static uint16_t update_attribute_handle(struct gatt_db_service *service,
 	previous_handle = service->attributes[index - 1]->handle;
 	service->attributes[index]->handle = previous_handle + 1;
 
-	return service->attributes[index]->handle;
+	return service->attributes[index];
 }
 
 static void set_attribute_data(struct gatt_db_attribute *attribute,
@@ -265,27 +300,28 @@ static void set_attribute_data(struct gatt_db_attribute *attribute,
 	attribute->user_data = user_data;
 }
 
-uint16_t gatt_db_add_characteristic(struct gatt_db *db, uint16_t handle,
-						const bt_uuid_t *uuid,
-						uint32_t permissions,
-						uint8_t properties,
-						gatt_db_read_t read_func,
-						gatt_db_write_t write_func,
-						void *user_data)
+struct gatt_db_attribute *
+gatt_db_service_add_characteristic(struct gatt_db_attribute *attrib,
+					const bt_uuid_t *uuid,
+					uint32_t permissions,
+					uint8_t properties,
+					gatt_db_read_t read_func,
+					gatt_db_write_t write_func,
+					void *user_data)
 {
-	uint8_t value[MAX_CHAR_DECL_VALUE_LEN];
 	struct gatt_db_service *service;
+	uint8_t value[MAX_CHAR_DECL_VALUE_LEN];
 	uint16_t len = 0;
 	int i;
 
-	service = queue_find(db->services, match_service_by_handle,
-							INT_TO_PTR(handle));
-	if (!service)
-		return 0;
+	if (!attrib)
+		return NULL;
+
+	service = attrib->service;
 
 	i = get_attribute_index(service, 1);
 	if (!i)
-		return 0;
+		return NULL;
 
 	value[0] = properties;
 	len += sizeof(properties);
@@ -294,98 +330,99 @@ uint16_t gatt_db_add_characteristic(struct gatt_db *db, uint16_t handle,
 	len += sizeof(uint16_t);
 	len += uuid_to_le(uuid, &value[3]);
 
-	service->attributes[i] = new_attribute(&characteristic_uuid, value,
-									len);
+	service->attributes[i] = new_attribute(service, &characteristic_uuid,
+								value, len);
 	if (!service->attributes[i])
-		return 0;
+		return NULL;
 
-	update_attribute_handle(service, i++);
+	attribute_update(service, i++);
 
-	service->attributes[i] = new_attribute(uuid, NULL, 0);
+	service->attributes[i] = new_attribute(service, uuid, NULL, 0);
 	if (!service->attributes[i]) {
 		free(service->attributes[i - 1]);
-		return 0;
+		return NULL;
 	}
 
 	set_attribute_data(service->attributes[i], read_func, write_func,
 							permissions, user_data);
 
-	return update_attribute_handle(service, i);
+	return attribute_update(service, i);
 }
 
-uint16_t gatt_db_add_char_descriptor(struct gatt_db *db, uint16_t handle,
-						const bt_uuid_t *uuid,
-						uint32_t permissions,
-						gatt_db_read_t read_func,
-						gatt_db_write_t write_func,
-						void *user_data)
+struct gatt_db_attribute *
+gatt_db_service_add_descriptor(struct gatt_db_attribute *attrib,
+					const bt_uuid_t *uuid,
+					uint32_t permissions,
+					gatt_db_read_t read_func,
+					gatt_db_write_t write_func,
+					void *user_data)
 {
 	struct gatt_db_service *service;
 	int i;
 
-	service = queue_find(db->services, match_service_by_handle,
-							INT_TO_PTR(handle));
-	if (!service)
-		return 0;
+	if (!attrib)
+		return false;
+
+	service = attrib->service;
 
 	i = get_attribute_index(service, 0);
 	if (!i)
-		return 0;
+		return NULL;
 
-	service->attributes[i] = new_attribute(uuid, NULL, 0);
+	service->attributes[i] = new_attribute(service, uuid, NULL, 0);
 	if (!service->attributes[i])
-		return 0;
+		return NULL;
 
 	set_attribute_data(service->attributes[i], read_func, write_func,
 							permissions, user_data);
 
-	return update_attribute_handle(service, i);
+	return attribute_update(service, i);
 }
 
-uint16_t gatt_db_add_included_service(struct gatt_db *db, uint16_t handle,
-						uint16_t included_handle)
+struct gatt_db_attribute *
+gatt_db_service_add_included(struct gatt_db_attribute *attrib,
+					struct gatt_db_attribute *include)
 {
-	struct gatt_db_service *included_service;
+	struct gatt_db_service *service, *included;
 	uint8_t value[MAX_INCLUDED_VALUE_LEN];
-	uint16_t len = 0;
-	struct gatt_db_service *service;
+	uint16_t included_handle, len = 0;
 	int index;
 
-	service = queue_find(db->services, match_service_by_handle,
-							INT_TO_PTR(handle));
-	if (!service)
-		return 0;
+	if (!attrib || !include)
+		return NULL;
 
-	included_service = queue_find(db->services, match_service_by_handle,
-						INT_TO_PTR(included_handle));
+	service = attrib->service;
+	included = include->service;
 
-	if (!included_service)
-		return 0;
+	/* Adjust include to point to the first attribute */
+	if (include != included->attributes[0])
+		include = included->attributes[0];
+
+	included_handle = include->handle;
 
 	put_le16(included_handle, &value[len]);
 	len += sizeof(uint16_t);
 
-	put_le16(included_handle + included_service->num_handles - 1,
-								&value[len]);
+	put_le16(included_handle + included->num_handles - 1, &value[len]);
 	len += sizeof(uint16_t);
 
 	/* The Service UUID shall only be present when the UUID is a 16-bit
 	 * Bluetooth UUID. Vol 2. Part G. 3.2
 	 */
-	if (included_service->attributes[0]->value_len == sizeof(uint16_t)) {
-		memcpy(&value[len], included_service->attributes[0]->value,
-				included_service->attributes[0]->value_len);
-		len += included_service->attributes[0]->value_len;
+	if (include->value_len == sizeof(uint16_t)) {
+		memcpy(&value[len], include->value, include->value_len);
+		len += include->value_len;
 	}
 
 	index = get_attribute_index(service, 0);
 	if (!index)
-		return 0;
+		return NULL;
 
-	service->attributes[index] = new_attribute(&included_service_uuid,
-								value, len);
+	service->attributes[index] = new_attribute(service,
+							&included_service_uuid,
+							value, len);
 	if (!service->attributes[index])
-		return 0;
+		return NULL;
 
 	/* The Attribute Permissions shall be read only and not require
 	 * authentication or authorization. Vol 2. Part G. 3.2
@@ -394,20 +431,15 @@ uint16_t gatt_db_add_included_service(struct gatt_db *db, uint16_t handle,
 	 */
 	set_attribute_data(service->attributes[index], NULL, NULL, 0, NULL);
 
-	return update_attribute_handle(service, index);
+	return attribute_update(service, index);
 }
 
-bool gatt_db_service_set_active(struct gatt_db *db, uint16_t handle,
-								bool active)
+bool gatt_db_service_set_active(struct gatt_db_attribute *attrib, bool active)
 {
-	struct gatt_db_service *service;
-
-	service = queue_find(db->services, match_service_by_handle,
-							INT_TO_PTR(handle));
-	if (!service)
+	if (!attrib)
 		return false;
 
-	service->active = active;
+	attrib->service->active = active;
 
 	return true;
 }
@@ -425,6 +457,7 @@ static void read_by_group_type(void *data, void *user_data)
 {
 	struct read_by_group_type_data *search_data = user_data;
 	struct gatt_db_service *service = data;
+	uint16_t grp_start, grp_end;
 
 	if (!service->active)
 		return;
@@ -436,7 +469,15 @@ static void read_by_group_type(void *data, void *user_data)
 	if (bt_uuid_cmp(&search_data->uuid, &service->attributes[0]->uuid))
 		return;
 
-	if (service->attributes[0]->handle < search_data->start_handle)
+	grp_start = service->attributes[0]->handle;
+	grp_end = grp_start + service->num_handles - 1;
+
+	if (grp_end < search_data->start_handle ||
+				grp_start > search_data->end_handle)
+		return;
+
+	if (service->attributes[0]->handle < search_data->start_handle ||
+		service->attributes[0]->handle > search_data->end_handle)
 		return;
 
 	/* Remember size of uuid */
@@ -449,8 +490,7 @@ static void read_by_group_type(void *data, void *user_data)
 		return;
 	}
 
-	queue_push_tail(search_data->queue,
-			UINT_TO_PTR(service->attributes[0]->handle));
+	queue_push_tail(search_data->queue, service->attributes[0]);
 }
 
 void gatt_db_read_by_group_type(struct gatt_db *db, uint16_t start_handle,
@@ -500,8 +540,7 @@ static void find_by_type(void *data, void *user_data)
 		if (bt_uuid_cmp(&search_data->uuid, &attribute->uuid))
 			continue;
 
-		queue_push_tail(search_data->queue,
-						UINT_TO_PTR(attribute->handle));
+		queue_push_tail(search_data->queue, attribute);
 	}
 }
 
@@ -551,8 +590,7 @@ static void read_by_type(void *data, void *user_data)
 		if (bt_uuid_cmp(&search_data->uuid, &attribute->uuid))
 			continue;
 
-		queue_push_tail(search_data->queue,
-						UINT_TO_PTR(attribute->handle));
+		queue_push_tail(search_data->queue, attribute);
 	}
 }
 
@@ -603,8 +641,7 @@ static void find_information(void *data, void *user_data)
 		if (attribute->handle > search_data->end_handle)
 			return;
 
-		queue_push_tail(search_data->queue,
-						UINT_TO_PTR(attribute->handle));
+		queue_push_tail(search_data->queue, attribute);
 	}
 }
 
@@ -624,7 +661,7 @@ void gatt_db_find_information(struct gatt_db *db, uint16_t start_handle,
 static bool find_service_for_handle(const void *data, const void *user_data)
 {
 	const struct gatt_db_service *service = data;
-	uint16_t handle = PTR_TO_INT(user_data);
+	uint16_t handle = PTR_TO_UINT(user_data);
 	uint16_t start, end;
 
 	start = service->attributes[0]->handle;
@@ -633,117 +670,20 @@ static bool find_service_for_handle(const void *data, const void *user_data)
 	return (start <= handle) && (handle < end);
 }
 
-bool gatt_db_read(struct gatt_db *db, uint16_t handle, uint16_t offset,
-				uint8_t att_opcode, bdaddr_t *bdaddr,
-				uint8_t **value, int *length)
-{
-	struct gatt_db_service *service;
-	uint16_t service_handle;
-	struct gatt_db_attribute *a;
-
-	if (!value || !length)
-		return false;
-
-	service = queue_find(db->services, find_service_for_handle,
-						INT_TO_PTR(handle));
-	if (!service)
-		return false;
-
-	service_handle = service->attributes[0]->handle;
-
-	a = service->attributes[handle - service_handle];
-	if (!a)
-		return false;
-
-	/*
-	 * We call callback, and set length to -1, to notify user that callback
-	 * has been called. Otherwise we set length to value length in database.
-	 */
-	if (a->read_func) {
-		*value = NULL;
-		*length = -1;
-		a->read_func(handle, offset, att_opcode, bdaddr, a->user_data);
-	} else {
-		if (offset > a->value_len)
-			return false;
-
-		*value = &a->value[offset];
-		*length = a->value_len - offset;
-	}
-
-	return true;
-}
-
-bool gatt_db_write(struct gatt_db *db, uint16_t handle, uint16_t offset,
-					const uint8_t *value, size_t len,
-					uint8_t att_opcode, bdaddr_t *bdaddr)
-{
-	struct gatt_db_service *service;
-	uint16_t service_handle;
-	struct gatt_db_attribute *a;
-
-	service = queue_find(db->services, find_service_for_handle,
-						INT_TO_PTR(handle));
-	if (!service)
-		return false;
-
-	service_handle = service->attributes[0]->handle;
-
-	a = service->attributes[handle - service_handle];
-	if (!a || !a->write_func)
-		return false;
-
-	a->write_func(handle, offset, value, len, att_opcode, bdaddr,
-								a->user_data);
-
-	return true;
-}
-
-const bt_uuid_t *gatt_db_get_attribute_type(struct gatt_db *db,
+struct gatt_db_attribute *gatt_db_get_attribute(struct gatt_db *db,
 							uint16_t handle)
 {
 	struct gatt_db_service *service;
-	struct gatt_db_attribute *attribute;
 	uint16_t service_handle;
 
-	service = queue_find(db->services, find_service_for_handle,
-						INT_TO_PTR(handle));
-	if (!service)
+	if (!db || !handle)
 		return NULL;
 
-	service_handle = service->attributes[0]->handle;
-
-	attribute = service->attributes[handle - service_handle];
-	if (!attribute)
+	service = queue_find(db->services, find_service_for_handle,
+							UINT_TO_PTR(handle));
+	if (!service)
 		return NULL;
 
-	return &attribute->uuid;
-}
-
-uint16_t gatt_db_get_end_handle(struct gatt_db *db, uint16_t handle)
-{
-	struct gatt_db_service *service;
-
-	service = queue_find(db->services, find_service_for_handle,
-						INT_TO_PTR(handle));
-	if (!service)
-		return 0;
-
-	return service->attributes[0]->handle + service->num_handles - 1;
-}
-
-bool gatt_db_get_attribute_permissions(struct gatt_db *db, uint16_t handle,
-							uint32_t *permissions)
-{
-	struct gatt_db_attribute *attribute;
-	struct gatt_db_service *service;
-	uint16_t service_handle;
-
-	service = queue_find(db->services, find_service_for_handle,
-							INT_TO_PTR(handle));
-	if (!service)
-		return false;
-
 	service_handle = service->attributes[0]->handle;
 
 	/*
@@ -751,11 +691,218 @@ bool gatt_db_get_attribute_permissions(struct gatt_db *db, uint16_t handle,
 	 * because find_service_for_handle() check if given handle is
 	 * in service range.
 	 */
-	attribute = service->attributes[handle - service_handle];
-	if (!attribute)
+	return service->attributes[handle - service_handle];
+}
+
+const bt_uuid_t *gatt_db_attribute_get_type(struct gatt_db_attribute *attrib)
+{
+	if (!attrib)
+		return NULL;
+
+	return &attrib->uuid;
+}
+
+uint16_t gatt_db_attribute_get_handle(struct gatt_db_attribute *attrib)
+{
+	if (!attrib)
+		return 0;
+
+	return attrib->handle;
+}
+
+bool gatt_db_attribute_get_service_uuid(struct gatt_db_attribute *attrib,
+							bt_uuid_t *uuid)
+{
+	struct gatt_db_service *service;
+
+	if (!attrib || !uuid)
 		return false;
 
-	*permissions = attribute->permissions;
-	return true;
+	service = attrib->service;
 
+	if (service->attributes[0]->value_len == sizeof(uint16_t)) {
+		uint16_t value;
+
+		value = get_le16(service->attributes[0]->value);
+		bt_uuid16_create(uuid, value);
+
+		return true;
+	}
+
+	if (service->attributes[0]->value_len == sizeof(uint128_t)) {
+		uint128_t value;
+
+		bswap_128(service->attributes[0]->value, &value);
+		bt_uuid128_create(uuid, value);
+
+		return true;
+	}
+
+	return false;
+}
+
+bool gatt_db_attribute_get_service_handles(struct gatt_db_attribute *attrib,
+						uint16_t *start_handle,
+						uint16_t *end_handle)
+{
+	struct gatt_db_service *service;
+
+	if (!attrib)
+		return false;
+
+	service = attrib->service;
+
+	if (start_handle)
+		*start_handle = service->attributes[0]->handle;
+
+	if (end_handle)
+		*end_handle = service->attributes[0]->handle +
+						service->num_handles - 1;
+
+	return true;
+}
+
+bool gatt_db_attribute_get_permissions(struct gatt_db_attribute *attrib,
+							uint32_t *permissions)
+{
+	if (!attrib || !permissions)
+		return false;
+
+	*permissions = attrib->permissions;
+
+	return true;
+}
+
+bool gatt_db_attribute_read(struct gatt_db_attribute *attrib, uint16_t offset,
+				uint8_t opcode, bdaddr_t *bdaddr,
+				gatt_db_attribute_read_t func, void *user_data)
+{
+	uint8_t *value;
+
+	if (!attrib || !func)
+		return false;
+
+	if (attrib->read_func) {
+		struct pending_read *p;
+
+		p = new0(struct pending_read, 1);
+		if (!p)
+			return false;
+
+		p->id = ++attrib->read_id;
+		p->func = func;
+		p->user_data = user_data;
+
+		queue_push_tail(attrib->pending_reads, p);
+
+		attrib->read_func(attrib, p->id, offset, opcode, bdaddr,
+							attrib->user_data);
+		return true;
+	}
+
+	/* Check boundary if value is stored in the db */
+	if (offset > attrib->value_len)
+		return false;
+
+	/* Guard against invalid access if offset equals to value length */
+	value = offset == attrib->value_len ? NULL : &attrib->value[offset];
+
+	func(attrib, 0, value, attrib->value_len - offset, user_data);
+
+	return true;
+}
+
+static bool find_pending(const void *a, const void *b)
+{
+	const struct pending_read *p = a;
+	unsigned int id = PTR_TO_UINT(b);
+
+	return p->id == id;
+}
+
+bool gatt_db_attribute_read_result(struct gatt_db_attribute *attrib,
+					unsigned int id, int err,
+					const uint8_t *value, size_t length)
+{
+	struct pending_read *p;
+
+	if (!attrib || !id)
+		return false;
+
+	p = queue_remove_if(attrib->pending_reads, find_pending,
+							UINT_TO_PTR(id));
+	if (!p)
+		return false;
+
+	p->func(attrib, err, value, length, p->user_data);
+
+	free(p);
+
+	return true;
+}
+
+bool gatt_db_attribute_write(struct gatt_db_attribute *attrib, uint16_t offset,
+					const uint8_t *value, size_t len,
+					uint8_t opcode, bdaddr_t *bdaddr,
+					gatt_db_attribute_write_t func,
+					void *user_data)
+{
+	if (!attrib || !func)
+		return false;
+
+	if (attrib->write_func) {
+		struct pending_write *p;
+
+		p = new0(struct pending_write, 1);
+		if (!p)
+			return false;
+
+		p->id = ++attrib->write_id;
+		p->func = func;
+		p->user_data = user_data;
+
+		queue_push_tail(attrib->pending_writes, p);
+
+		attrib->write_func(attrib, p->id, offset, value, len, opcode,
+						bdaddr, attrib->user_data);
+		return true;
+	}
+
+	/* For values stored in db allocate on demand */
+	if (!attrib->value || offset >= attrib->value_len ||
+				len > (unsigned) (attrib->value_len - offset)) {
+		attrib->value = realloc(attrib->value, len + offset);
+		if (!attrib->value)
+			return false;
+		/* Init data in the first allocation */
+		if (!attrib->value_len)
+			memset(attrib->value, 0, offset);
+		attrib->value_len = len + offset;
+	}
+
+	memcpy(&attrib->value[offset], value, len);
+
+	func(attrib, 0, user_data);
+
+	return true;
+}
+
+bool gatt_db_attribute_write_result(struct gatt_db_attribute *attrib,
+						unsigned int id, int err)
+{
+	struct pending_write *p;
+
+	if (!attrib || !id)
+		return false;
+
+	p = queue_remove_if(attrib->pending_writes, find_pending,
+							UINT_TO_PTR(id));
+	if (!p)
+		return false;
+
+	p->func(attrib, err, p->user_data);
+
+	free(p);
+
+	return true;
 }
